/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [pl0_int.c]                Language Interface                            */
/*                                                                          */
/* ------------------------------------------------------------------------ */

#include "ptm.h"
#include "gls.h"

/*I--------------------- symbol objects - init & quit --------------------- */

static symbol* CfgSyms = (symbol*)NULL;
static symbol* PrdSyms = (symbol*)NULL;
static int initcount_pl0 = 0;

void pl0_initSymbols()
{
  if( initcount_pl0 == 0 )
  {
    GLS_init();
    assert0(CfgSyms == (symbol*)NULL,"pl0Symbols already initialized");
    CfgSyms = (symbol*)NewMem(sizeof(symbol)*35);
    PrdSyms = (symbol*)NewMem(sizeof(symbol)*36);

    CfgSyms[0] = stringToSymbol("Tok");
    CfgSyms[1] = stringToSymbol("Int");
    CfgSyms[2] = stringToSymbol("Ide");
    CfgSyms[3] = stringToSymbol("Com");
    CfgSyms[4] = stringToSymbol("fun");
    CfgSyms[5] = stringToSymbol("(");
    CfgSyms[6] = stringToSymbol(",");
    CfgSyms[7] = stringToSymbol(")");
    CfgSyms[8] = stringToSymbol("=");
    CfgSyms[9] = stringToSymbol("run");
    CfgSyms[10] = stringToSymbol("if");
    CfgSyms[11] = stringToSymbol("then");
    CfgSyms[12] = stringToSymbol("else");
    CfgSyms[13] = stringToSymbol("<");
    CfgSyms[14] = stringToSymbol("+");
    CfgSyms[15] = stringToSymbol("-");
    CfgSyms[16] = stringToSymbol("*");
    CfgSyms[17] = stringToSymbol("/");
    CfgSyms[18] = stringToSymbol("Program");
    CfgSyms[19] = stringToSymbol("_List1");
    CfgSyms[20] = stringToSymbol("_List1_0");
    CfgSyms[21] = stringToSymbol("_List2");
    CfgSyms[22] = stringToSymbol("_List2_0");
    CfgSyms[23] = stringToSymbol("Dfn");
    CfgSyms[24] = stringToSymbol("_List3");
    CfgSyms[25] = stringToSymbol("_List3_0");
    CfgSyms[26] = stringToSymbol("Run");
    CfgSyms[27] = stringToSymbol("Exp");
    CfgSyms[28] = stringToSymbol("Exp1");
    CfgSyms[29] = stringToSymbol("Exp2");
    CfgSyms[30] = stringToSymbol("Exp3");
    CfgSyms[31] = stringToSymbol("Exp4");
    CfgSyms[32] = stringToSymbol("_List4");
    CfgSyms[33] = stringToSymbol("_List4_0");
    CfgSyms[34] = stringToSymbol("pl0");

    PrdSyms[0] = stringToSymbol("pgm");
    PrdSyms[1] = stringToSymbol("nil");
    PrdSyms[2] = stringToSymbol("cons");
    PrdSyms[3] = stringToSymbol("nil");
    PrdSyms[4] = stringToSymbol("cons");
    PrdSyms[5] = stringToSymbol("nil");
    PrdSyms[6] = stringToSymbol("cons");
    PrdSyms[7] = stringToSymbol("nil");
    PrdSyms[8] = stringToSymbol("cons");
    PrdSyms[9] = stringToSymbol("fun");
    PrdSyms[10] = stringToSymbol("nil");
    PrdSyms[11] = stringToSymbol("cons");
    PrdSyms[12] = stringToSymbol("nil");
    PrdSyms[13] = stringToSymbol("cons");
    PrdSyms[14] = stringToSymbol("run");
    PrdSyms[15] = stringToSymbol("ign0");
    PrdSyms[16] = stringToSymbol("if");
    PrdSyms[17] = stringToSymbol("ign0");
    PrdSyms[18] = stringToSymbol("les");
    PrdSyms[19] = stringToSymbol("equ");
    PrdSyms[20] = stringToSymbol("ign0");
    PrdSyms[21] = stringToSymbol("add");
    PrdSyms[22] = stringToSymbol("sub");
    PrdSyms[23] = stringToSymbol("ign0");
    PrdSyms[24] = stringToSymbol("mlt");
    PrdSyms[25] = stringToSymbol("div");
    PrdSyms[26] = stringToSymbol("neg");
    PrdSyms[27] = stringToSymbol("ign0");
    PrdSyms[28] = stringToSymbol("int");
    PrdSyms[29] = stringToSymbol("var");
    PrdSyms[30] = stringToSymbol("app");
    PrdSyms[31] = stringToSymbol("nil");
    PrdSyms[32] = stringToSymbol("cons");
    PrdSyms[33] = stringToSymbol("nil");
    PrdSyms[34] = stringToSymbol("cons");
    PrdSyms[35] = stringToSymbol("Start_Program");
  }
  ++initcount_pl0;
}

void pl0_quitSymbols()
{
  --initcount_pl0;
  if( initcount_pl0 == 0 )
  {
    assert0(CfgSyms != (symbol*)NULL,"pl0Symbols not initialized");
    if( CfgSyms != (symbol*)NULL ) FreeMem(CfgSyms);
    CfgSyms = (symbol*)NULL;
    if( PrdSyms != (symbol*)NULL ) FreeMem(PrdSyms);
    PrdSyms = (symbol*)NULL;
  }
}

/*I-------------------------- Types & Constants --------------------------- */

AbstractType( pl0 );

AbstractType( pl0Program );
AbstractType( pl0Dfn );
AbstractType( pl0Run );
AbstractType( pl0Exp );

/*I--------------------------- Access to Tokens --------------------------- */

c_bool Tpl0_Int(GLS_Tok x)
{
  return( PT_THISTOK((PT_Term)x, CfgSyms[1]) );
}

c_bool Tpl0_Ide(GLS_Tok x)
{
  return( PT_THISTOK((PT_Term)x, CfgSyms[2]) );
}

/*I--------------------------- Access to Terms ---------------------------- */

c_bool pl0_pl0(PT_Term x, pl0* x1)
{
  if( PT_THISNTM(x, CfgSyms[34]) )
  {
    if( (StdCPtr)x1 != (StdCPtr)NULL ) *x1 = (pl0)x;
    return True;
  }
  return False;
}

c_bool pl0_Program(PT_Term x, pl0Program* x1)
{
  if(  PT_THISNTM(x, CfgSyms[18]) )
  {
    if( (StdCPtr)x1 != (StdCPtr)NULL ) *x1 = (pl0Program)x;
    return True;
  }
  return False;
}

c_bool pl0_Dfn(PT_Term x, pl0Dfn* x1)
{
  if(  PT_THISNTM(x, CfgSyms[23]) )
  {
    if( (StdCPtr)x1 != (StdCPtr)NULL ) *x1 = (pl0Dfn)x;
    return True;
  }
  return False;
}

c_bool pl0_Run(PT_Term x, pl0Run* x1)
{
  if(  PT_THISNTM(x, CfgSyms[26]) )
  {
    if( (StdCPtr)x1 != (StdCPtr)NULL ) *x1 = (pl0Run)x;
    return True;
  }
  return False;
}

c_bool pl0_Exp(PT_Term x, pl0Exp* x1)
{
  if(  PT_THISNTM(x, CfgSyms[27])
    || PT_THISNTM(x, CfgSyms[31])
    || PT_THISNTM(x, CfgSyms[28])
    || PT_THISNTM(x, CfgSyms[29])
    || PT_THISNTM(x, CfgSyms[30]) )
  {
    if( (StdCPtr)x1 != (StdCPtr)NULL ) *x1 = (pl0Exp)x;
    return True;
  }
  return False;
}

/*I--------------------------------- pl0 ---------------------------------- */

c_bool pl0_Start_Program(pl0 x, pl0Program* x1)
#define pl0_Start_0   pl0_Start_Program
{
  assert0( PT_THISNTM((PT_Term)x, CfgSyms[34]), "pl0 expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[35]) ) return False;
  if( (StdCPtr)x1 != (StdCPtr)NULL )
    *x1 = (pl0Program)GLS_ign_Skip(PT_nthSymbol((PT_Term)x, CfgSyms[18], 1, True));
  return True;
}

/*I------------------------------- Program -------------------------------- */

c_bool pl0Program_pgm(pl0Program x, GLS_Lst(pl0Dfn)* x1, GLS_Lst(pl0Run)* x2)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[18]), "Program expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[0]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Program)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Lst(pl0Dfn))GLS_ign_Skip(x);
      x = (pl0Program)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (GLS_Lst(pl0Run))GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Program,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Lst(pl0Dfn))GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Program,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (GLS_Lst(pl0Run))GLS_ign_Skip(y);
    }
  }
  return True;
}

/*I--------------------------------- Dfn ---------------------------------- */

c_bool pl0Dfn_fun(pl0Dfn x, GLS_Tok* x1, GLS_Lst(GLS_Tok)* x2,
                  pl0Exp* x3)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[23]), "Dfn expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[9]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Dfn)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Tok)GLS_ign_Skip(x);
      x = (pl0Dfn)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (GLS_Lst(GLS_Tok))GLS_ign_Skip(x);
      x = (pl0Dfn)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x3 != (StdCPtr)NULL )
        *x3 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Dfn,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Tok)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Dfn,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (GLS_Lst(GLS_Tok))GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Dfn,x);
      if( (StdCPtr)x3 != (StdCPtr)NULL )
        *x3 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

/*I--------------------------------- Run ---------------------------------- */

c_bool pl0Run_run(pl0Run x, pl0Exp* x1)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[26]), "Run expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[14]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Run)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Run,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

/*I--------------------------------- Exp ---------------------------------- */

c_bool pl0Exp_add(pl0Exp x, pl0Exp* x1, pl0Exp* x2)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[21]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_mlt(pl0Exp x, pl0Exp* x1, pl0Exp* x2)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[24]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_neg(pl0Exp x, pl0Exp* x1)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[26]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_app(pl0Exp x, GLS_Tok* x1, GLS_Lst(pl0Exp)* x2)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[30]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Tok)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (GLS_Lst(pl0Exp))GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Tok)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (GLS_Lst(pl0Exp))GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_sub(pl0Exp x, pl0Exp* x1, pl0Exp* x2)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[22]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_les(pl0Exp x, pl0Exp* x1, pl0Exp* x2)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[18]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_if(pl0Exp x, pl0Exp* x1, pl0Exp* x2,
                 pl0Exp* x3)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[16]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x3 != (StdCPtr)NULL )
        *x3 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x3 != (StdCPtr)NULL )
        *x3 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_int(pl0Exp x, GLS_Tok* x1)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[28]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Tok)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Tok)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_div(pl0Exp x, pl0Exp* x1, pl0Exp* x2)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[25]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_var(pl0Exp x, GLS_Tok* x1)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[29]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Tok)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (GLS_Tok)GLS_ign_Skip(y);
    }
  }
  return True;
}

c_bool pl0Exp_equ(pl0Exp x, pl0Exp* x1, pl0Exp* x2)
{
  assert0(  PT_THISNTM((PT_Term)x, CfgSyms[27])
        || PT_THISNTM((PT_Term)x, CfgSyms[31])
        || PT_THISNTM((PT_Term)x, CfgSyms[28])
        || PT_THISNTM((PT_Term)x, CfgSyms[29])
        || PT_THISNTM((PT_Term)x, CfgSyms[30]), "Exp expected" );
  if( ! PT_THISPRD((PT_Term)x, PrdSyms[19]) ) return False;
  { PT_Term y;
    if( ! PT_isXaron(x) )
    {
      x = (pl0Exp)GLS_keycom_Skip(PT_parts(x));
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(x);
      x = (pl0Exp)GLS_keycom_Skip(PT_next(x));
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(x);
    }
    else
    {
      x = XGLS_keycom_Skip(XPT_parts(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x1 != (StdCPtr)NULL )
        *x1 = (pl0Exp)GLS_ign_Skip(y);
      x = XGLS_keycom_Skip(rst(x)); y = fst(pl0Exp,x);
      if( (StdCPtr)x2 != (StdCPtr)NULL )
        *x2 = (pl0Exp)GLS_ign_Skip(y);
    }
  }
  return True;
}
